;; todo:
;;   sysenter
;;   sleep / timer int wakeup
;;   -move codepage to autocfg / internationalize kbd


%include "autocfg.inc"
%include "sos.inc"


;______________________________________________________________________________
;
; boot sector - load rest of os, init hardware, enter paged pmode
;______________________________________________________________________________
;
section boot use16

; init mouse
	mov	al, 0xa8			; enable kbd aux device
	out	0x64, al
	mov	ax, 0xc203			; set mouse sensitivity
	mov	bh, 0x03			; to 8 counts / mm
	int	0x15
	mov	al, 0xd4
	out	0x64, al
	mov	al, 0xf4			; ps/2 enable cmd
	out	0x60, al

; load image
	mov	ax, 0x0224			; read 36 sectors
	cwd					; from disk
	mov	es, dx				;
	mov	bx, 0x1000			; into mem at 0x01000 phys
	mov	cx, 0x0002			; starting with sector 2
	int	0x13

; set vbe mode
	mov	ax, 0x4f02
	mov	bx, VMODE|0xc800		; lfb, no mem clear, use crtc pll timing data
	mov	di, crtc_timing
	int	0x10

; enable a20, init pic, load idtr&gdtr, enter pmode
	mov	ah, 0x89
	mov	bx, 0x2028			; irq 0x00..0x0f -> int 0x20..0x2f
	mov	si, gdt
	int	0x15
bits 32

; load tr
; > ax = 0x00??
	mov	al, TSS
	ltr	ax
	mov	byte [tss_desc+5], 0x89		; (ltr sets busy flag)

; init page dir
	xor	eax, eax
	xor	ecx, ecx
	mov	ch, 3
	xor	edi, edi
rep	stosd					; clear pdes for 0..3 gb
	mov	al, 0x83
	mov	[ecx], al			; self map 0..4 mb (temporary)
	mov	ah, 1				; global
	stosd					; os pde (0..4 mb mapped to 0xc0000000)
	mov	eax, 0xc0400083
	stosd
	add	eax, 0x00400000
	jnc	$-6				; self map 3..4 gb

; enable paging w/ 4 mb and global pages
; > eax = 0x000000??
	mov	al, 0x10			; pse = 1
	mov	cr4, eax
	mov	edi, 0x80000021
	mov	cr0, edi
	mov	al, 0x90			; pse = 1, pge = 1; pge must not be set w/o paging enabled
	mov	cr4, eax

; init sysenter
; > eax = 0x000000??
; > ecx = 0x0000????
	mov	cx, 0x174
	mov	al, CODE
	cdq					; edx = 0
	wrmsr					; SYSENTER_CS_MSR
	inc	ecx
	mov	ax, stack_top			; < 64 kb
	mov	esp, eax
	wrmsr					; SYSENTER_ESP_MSR
	inc	ecx
	mov	eax, api_entry
	wrmsr					; SYSENTER_EIP_MSR

; set lfb cache type to wc via mtrr
; > ecx = 0x0000????
; > edx = 0
	mov	cx, 0x20e
	mov	eax, lfb|1			; base address; type = wc
	lea	edi, [eax-1]
	wrmsr					; MTRRPhysBase7
	inc	ecx
	mov	eax, (-lfb_size)|0x800		; length of range; valid bit = 1
	wrmsr

; clear screen
; > edi -> lfb
	mov	eax, 0x00396da5			; light blue
	mov	ecx, XRES*YRES
rep	stosd

; clear pic imr (all ints disabled by bios int 0x15 func 0x89)
	xor	eax, eax
	out	0x21, al
	out	0xa1, al

; turn off floppy motor
; > al = 0
	mov	dx, 0x3f2
	out	dx, al

; gfx card specific (included in autocfg.inc)
; > eax = 0
; > ecx = 0
; > dx = 0x03??
init_gfx

; set pit channel 0 to 100 hz
	mov	al, 0x9c
	out	0x40, al
	mov	al, 0x2e
	out	0x40, al


	sti






mov ecx,100
mov edi,lfb+4096*100+100*4
mov eax,-1
rep stosd


call bigtest




stop:
nop
nop
nop

jmp stop








char db 0,0

crtc_timing:
incbin "crtc.dat"

times 510+$$-$	db	0x00
		dw	0xaa55


section code use32 align=0x10





bigtest:
cli
xor eax,eax
mov byte [eax+4],0x87
mov cr3,eax
jmp USER_CODE:usertest

ret

usertest equ $+0x400000
	sysenter
_usercont equ $+0x400000
	sysenter
_usercont2 equ $+0x400000
	jmp $


api_entry:
	push	ds
	push	es
	push	ss
	push	ss
	pop	ds
	pop	es
	mov	edi, [xpos]
	lea	edi, [edi*4+lfb+4096*600]
	mov	ecx, 300
	mov	eax, 0x00000000
rep	stosd
	add	dword [xpos], 600
	mov	edx, [usercont]
	mov	ecx, stack_top-64
	mov	dword [usercont], _usercont2
	pop	es
	pop	ds
	sysexit

[section data]
usercont dd _usercont
xpos dd 100
__SECT__



dump_edx:
	mov	ecx, 8
charloop:
	rol	edx, 4
	mov	eax, edx
	and	eax, 0x0f
	cmp	eax, 9
	ja	.1
	add	eax, '0'
	jmp	.2
.1:	add	eax, 'A'-10
.2:	mov	[char], al
	mov	ebx, char
	pusha
	call	draw_text
	popa
	add	dword [x_pos], 9
	loop	charloop

	add	dword [x_pos], 90
	cmp	dword [x_pos], 900
	jb	.3
	mov	dword [x_pos], 10
	add	dword [y_pos], 30
.3:

	ret











;______________________________________________________________________________
;
; kbd handler
;______________________________________________________________________________
;
isr 0x21
	pusha

mov edi,lfb+4096*600+100
mov eax,0x00ff0000
mov ecx,200
rep stosd


; get scancode / convert to ascii
	xor	eax, eax
	in	al, 0x60			; al = scancode | (key down)? 0 : 0x80
	btr	eax, 7				; al = scancode; cf = !key down
	setnc	[keys+eax]			; update "key currently down" array
	mov	ah, [keys+K_SHIFT]		; use shift table if shift pressed
	mov	ebx, [kbd_buf_pos]
	mov	cl, [ascii+eax]			; cl = ascii code

; add ascii code to kbd_buf
	mov	[kbd_buf+ebx], cl
	sbb	byte [kbd_buf_pos], -1		; if(key down) buf[pos++] = ascii

	mov	al, 0x20
	out	0x20, al
	popa
	iret


;______________________________________________________________________________
;
; ps/2 mouse handler
;______________________________________________________________________________
;
isr 0x2c
	pusha

; store mouse data packet as 0x00bbxxyy
	shl	dword [packet], 8
	in	al, 0x60			; al = data from mouse
	mov	[packet], al
	jnc	.ret				; all 3 bytes received? no - exit

; buttons

; scale delta if abs(delta) > threshold
	movsx	eax, al				; dy
	movsx	ebx, byte [packet+1]		; dx
	xor	esi, esi
	mov	edi, [threshold]
	lea	ecx, [eax*4+esi]		; (base reg avoids 32 bit offset)
	lea	edx, [ebx*4+esi]
	cmp	eax, edi
	cmovg	eax, ecx
	cmp	ebx, edi
	cmovg	ebx, edx
	neg	edi
	cmp	eax, edi
	cmovl	eax, ecx
	cmp	ebx, edi
	cmovl	ebx, edx

; clip & update position
; > esi = 0
	mov	ecx, [cursor_pos]		; old y pos << 16
	mov	edi, YRES << 16
	mov	ebp, XRES
	movsx	edx, cx				; old x pos
	shl	eax, 16
	sub	ecx, eax			; new y pos << 16
	cmovs	ecx, esi
	cmp	ecx, edi
	cmova	ecx, edi
	add	edx, ebx			; new x pos
	cmovs	edx, esi
	cmp	edx, ebp
	cmova	edx, ebp
	mov	cx, dx
	mov	[cursor_pos], ecx

	mov	dword [packet], 0x00000100	; reset

.ret:
	mov	al, 0x20
	out	0x20, al
	out	0xa0, al			; irq 0x0c (slave pic)
	popa
	iret





%if 0
;______________________________________________________________________________
;
; memory allocation (freed on program termination)
; > ecx -  # 4 mb pages
; < eax -> mem, valid if cf = 0
;______________________________________________________________________________
;
mmap:
	mov	edx, [pg_avail]
	cld					; safety
	mov	ebx, [cur_task]
	sub	edx, ecx			; edx = pages left
	mov	edi, [ebx+pcb_pg_dir]		; edi -> next free pde
	jbe	.ret				; not enough mem? yes - quit
	mov	[pg_avail], edx
	mov	eax, edi
	lea	esi, [pg_stack+edx*4]		; esi -> page stack
	shl	eax, 20				; eax = virtual addr
rep	movsd					; copy pages from stack to pd
	mov	[ebx+pcb_pg_dir], edi
.ret:
	sysret

; free mapped mem
	mov	ebx, [cur_task]
	mov	esi, [ebx+pcb_pg_dir]
	mov	edi, [pg_stack]
	mov	ecx, esi
	and	ecx, 0xfff
	sub	esi, ecx
	shr	ecx, 2
rep	movsd
	mov	[pg_stack], edi




; init pg dir
	mov	ebx, [cur_task]
	mov	edi, [ebx+pcb_pg_dir]



%endif





;______________________________________________________________________________
;
; text display
; regs:
;   x eax -  text color
;   > ebx -> asciiz string
;   x ecx
;   x edx -> last pos in lfb
;   x esi -> rle sprite for current char
;   x edi -> pos in lfb
;   x ebp -> font
;______________________________________________________________________________
;
draw_text:
; calc lfb offset
	mov	edx, [y_pos]
	mov	ecx, [x_pos]
	shl	edx, 12				; ***
	mov	eax, [text_color]
	lea	edi, [lfb+edx+ecx*4]		; ***

.skip_prologue:
	mov	ebp, [font]

.char_loop:
; esi -> rle sprite for current char
	movzx	ecx, byte [ebx]			; ecx = current char
	mov	esi, ebp			; esi -> font
	sub	ecx, byte 32
	js	.ret				; '\0' reached? yes - quit
	inc	ebx				; ebx -> next char
	add	esi, [ebp+ecx*4]		; esi -> rle sprite for current char

; save edi
	mov	edx, edi

; draw char
align 0x10
.draw_run:
	add	edi, [byte esi+0]		; trans len
	mov	ecx, [byte esi+4]		; solid len
	jc	.next_char			; end of char reached? yes - quit
	add	esi, byte 8
rep	stosd					; blit solid pixels
	jmp	.draw_run
.next_char:

; edi -> next pos in lfb
	lea	edi, [edx+8*4]			; ***

	jmp	.char_loop

.ret:
	ret



















inth1:
push eax
mov al,0x20
out 0x20,al
pop eax
iret

inth2:
push eax
mov al,0x20
out 0x20,al
out 0xa0,al
pop eax
iret









section data use32 align=0x10


;_______________________________________________________________________
;
; gdt/idt
;_______________________________________________________________________
;
gdt		equ	$-8			; skip null descriptor

; gdtr
		dw	0x0f47
		dd	gdt
		dw	0

; idtr
		dw	0x017f
		dd	idt
		dw	0

		desc 0, -1, D_DATA		; temp ds
		desc 0, -1, D_DATA		; temp es
		desc 0, -1, D_DATA		; temp ss
		desc 0, -1, D_CODE		; temp cs
		desc 0, 0, 0			; bios reserved

CODE		desc 0, -1, D_CODE		; sysenter needs these to be continuous
DATA		desc 0, -1, D_DATA
USER_CODE	desc 0, -1, D_CODE+D_DPL3
USER_DATA	desc 0, -1, D_DATA+D_DPL3

tss_desc:
TSS		desc tss, 0x67, D_TSS

idt:
; processor exceptions
%rep 0x20
		desc 0, CODE, D_TRAP
%endrep

; hardware interrupts
%rep 0x8
		desc inth1, CODE, D_INT
%endrep
%rep 0x8
		desc inth2, CODE, D_INT
%endrep


;______________________________________________________________________________
;
; mouse
;______________________________________________________________________________
;
packet		dd	0x00000100
threshold	dd	5

arrow_cursor:
incbin "arrow.dat"


;______________________________________________________________________________
;
; kbd
;______________________________________________________________________________
;
ascii:
		db	0, K_ESC
		db	'1234567890-='
		db	K_BACK, K_TAB
		db	'qwertyuiop[]'
		db	K_ENTER, K_CTRL
		db	"asdfghjkl;'"
		db	0, K_SHIFT
		db	'\zxcvbnm,./'
		db	K_SHIFT, '*', K_ALT, K_SPACE, 0
		db	K_F1, K_F2, K_F3, K_F4, K_F5
		db	K_F6, K_F7, K_F8, K_F9, K_F10
		db	0, 0
		db	'789-456+1230'
		db	K_DEL, 0, 0, 0, K_F11, K_F12		; 89 till end
		db	0, 0, 0, 0, 0, 0, 0	; alignment

keys		resb	0x80
kbd_buf		resb	0x80

; shift table = ascii table+0x100

		db	0, K_ESC
		db	'!@#$%^&*()_+'
		db	K_BACK, K_TAB
		db	'QWERTYUIOP{}'
		db	K_ENTER, K_CTRL
		db	'ASDFGHJKL:"~'
		db	K_SHIFT
		db	'|ZXCVBNM<>?'
		db	K_SHIFT, '*', K_ALT, K_SPACE, 0
		db	K_F1, K_F2, K_F3, K_F4, K_F5
		db	K_F6, K_F7, K_F8, K_F9, K_F10
		db	0, 0
		db	'789-456+1230'
		db	K_DEL, 0, 0, 0, K_F11, K_F12

kbd_buf_pos	dd	0			; index into kbd_buf



x_pos		dd	10			; ***
y_pos		dd	10			; ***

text_color	dd	0

align 0x10
sys8x16:
incbin "sys8x16.rle"

font		dd	sys8x16




tmp dd 0




section bss use32 align=0x10

tss		resb	0x68
		resb	8		; alignment
resb 0x100
stack_top:
